import java.util.Iterator;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.GregorianCalendar;


/**
 * A LogIterator is an Iterator as you already know it.
 * It provides the hasNext() and next() methods.
 */
public class LogIterator implements Iterator<LogEntry> {

    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss Z");
    private static final Pattern PATTERN = Pattern.compile("(\\S+) \\S+ \\S+ \\[(.*)\\] \"(.*)\" ([-0-9]+) ([-0-9]+)");

    private BufferedReader reader;
    private String currentLine;
    private boolean isOpen;
    
    
    /**
     * Create a LogIterator for the given log file name.
     * Use the Log class' iterator() method instead of calling this constructor.
     */
    public LogIterator(final String fileName) {
        try {
            reader = new BufferedReader(new FileReader(fileName));
            isOpen = true;
        } catch (final IOException ex) {
            ex.printStackTrace();
            reader = null;
            isOpen = false;
        }
        fetchLine();
    }
    
    private void fetchLine() {
        if (isOpen) {
            try {
                currentLine = reader.readLine();
            } catch (final IOException ex) {
                currentLine = null;
            }
            if (currentLine==null) {
                try {
                    reader.close();
                } catch (final IOException ex) {
                    // ignore
                }
                isOpen = false;
            }
        }
    }
    
    /**
     * Do we have at least one more LogEntry?
     */
    public boolean hasNext() {
        return currentLine!=null;
    }
    
    /**
     * Get a LogEntry and move to the next one.
     */
    public LogEntry next() {
        final LogEntry entry = parseLine(currentLine);
        fetchLine();
        return entry;
    }
    
    /**
     * This Iterator does not allow you to remove anything.
     */
    public void remove() {
        throw new UnsupportedOperationException();
    }
    
    private LogEntry parseLine(final String line) {
        Matcher m = PATTERN.matcher(line);
        if (m.matches()) {
            final String remoteHost = m.group(1);
            final TimeStamp timeStamp = parseTimeStamp(m.group(2));
            final String request = m.group(3);
            final int status = parseInt(m.group(4));
            final int size = parseInt(m.group(5));
            return new LogEntry(remoteHost, timeStamp, request, status, size);
        } else {
            System.err.println("The following line does not seem to match the common log file format:");
            System.err.println(line);
            return null;
        }
    }
    
    private TimeStamp parseTimeStamp(final String string) {
        final GregorianCalendar date = new GregorianCalendar();
        try {
            date.setTime(DATE_FORMAT.parse(string));
            return new TimeStamp(date);
        } catch (final ParseException ex) {
            System.err.println("The following date does not seem to be parseable: "+string);
            return null;
        }
    }
    
    private int parseInt(final String string) {
        if (string.equals("-")) {
            return 0;
        } else {
            return Integer.parseInt(string);
        }
    }
    
}
